/*
This file is a part of sleid0r -- slideshow viewer/generator.
Copyright (C) 2013-2014 Vadim Druzhin <cdslow@mail.ru>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <SDL.h>
#include <SDL_image.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <time.h>
#include "err.h"
#include "mod_frei0r.h"
#include "images.h"

typedef struct loop_check
    {
    struct stat st;
    struct loop_check const *parent;
    } loop_check_t;

static char **add_file(char **imgv, int *imgc, char const *path);
static char **add_dir(char **imgv, int *imgc, char const *dir, loop_check_t const *parent);
static int is_dir(char const *path, struct stat *st);
static int is_loop(loop_check_t const *lc);
static int is_image(char const *path);
static int cmp_files(const void *a, const void *b);
static void image_fastscale(
    SDL_Surface *dst, SDL_Surface *src, int w, int h, int dx, int dy, int km, int kd, int sx, int sy);
static void image_overscale(SDL_Surface *dst, SDL_Surface *src, int w, int h, int dx, int dy, int om, int od);
static void image_upscale(SDL_Surface *dst, SDL_Surface *src, int w, int h, int dx, int dy);
static void par(int a[3], int y1, int y2, int y3);
static inline void par32(int a[3 * 4], Uint32 px0, Uint32 px1, Uint32 px2);
static inline Uint32 get_pixel1(int const a[3 * 4], int dx, int div);
static inline Uint32 get_pixel2(int const a1[3 * 4], int const a2[3 * 4], int dx, int div);
static inline Uint32 get_color1(int const a[3], int dx, int div);
static inline Uint32 get_color2(int const a1[3], int const a2[3], int dx, int div);
static void image_downscale(SDL_Surface *dst, SDL_Surface *src, int w, int h, int dx, int dy);
static inline void add32(int sum[4], Uint32 px, int mul, int div);
static inline Uint32 div32(int const sum[4], int div);
static char *dir_name(char const *path);
static int dir_present(char **dirs, int dirc, char const *path);
static int dir_match(char const *dir, char const *path);

SDL_Surface *image_realloc(SDL_Surface *img, int w, int h)
    {
    SDL_PixelFormat fmt;
    SDL_VideoInfo const *vi;

    if(img != NULL)
        {
        if(img->w == w && img->h == h)
            return img;

        fmt = *img->format;

        SDL_FreeSurface(img);
        }
    else
        {
        vi = SDL_GetVideoInfo();

        if(NULL == vi)
            return image_alloc_rgba(w, h);

        if(vi->vfmt->BitsPerPixel != 32)
            {
            fprintf(stderr, "Bad video format (BitsPerPixel=%d)\n",
                vi->vfmt->BitsPerPixel);
            exit(EXIT_FAILURE);
            }

        fmt = *vi->vfmt;
        }

    return image_alloc_32(w, h, fmt.Rmask, fmt.Gmask, fmt.Bmask, fmt.Amask);
    }

SDL_Surface *image_alloc_rgba(int w, int h)
    {
    return image_alloc_32(w, h, MASK_R, MASK_G, MASK_B, MASK_A);
    }

SDL_Surface *image_alloc_bgra(int w, int h)
    {
    return image_alloc_32(w, h, MASK_B, MASK_G, MASK_R, MASK_A);
    }

SDL_Surface *image_alloc_32(int w, int h, Uint32 r, Uint32 g, Uint32 b, Uint32 a)
    {
    SDL_Surface *img;

    img = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, 32, r, g, b, a);

    if(NULL == img)
        {
        fprintf(stderr, "SDL_CreateRGBSurface failed: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
        }

    SDL_SetAlpha(img, 0, SDL_ALPHA_OPAQUE);

    return img;
    }

int image_copy(SDL_Surface *dst, SDL_Surface *src)
    {
    int ret = SDL_BlitSurface(src, NULL, dst, NULL);

    if(ret < 0)
        fprintf(stderr, "BlitSurface failed: %s\n", SDL_GetError());

    return ret;
    }

SDL_Surface *image_convert(SDL_Surface *src, SDL_PixelFormat *fmt)
    {
    SDL_Surface *img;

    img = SDL_ConvertSurface(src, fmt, SDL_SWSURFACE);

    if(NULL == img)
        {
        fprintf(stderr, "SDL_ConvertSurface failed: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
        }

    return img;
    }

char **image_list(int *imgc, char **argv)
    {
    *imgc = 0;

    return image_list_append(NULL, imgc, argv);
    }

char **image_list_append(char **imgv, int *imgc, char **argv)
    {
    loop_check_t lc;

    while(*argv != NULL)
        {
        if(is_dir(*argv, &lc.st))
            {
            lc.parent = NULL;
            imgv = add_dir(imgv, imgc, *argv, &lc);
            }
        else
            imgv = add_file(imgv, imgc, *argv);

        ++argv;
        }

    return imgv;
    }

void image_list_free(char **imgv, int imgc)
    {
    if(imgv != NULL)
        {
        for(--imgc; imgc >= 0; --imgc)
            {
            if(imgv[imgc] != NULL)
                free(imgv[imgc]);
            }
        free(imgv);
        }
    }

static char **add_file(char **imgv, int *imgc, char const *path)
    {
    imgv = e_realloc(imgv, (*imgc + 2) * sizeof(*imgv));

    imgv[*imgc] = e_strdup(path);

    ++*imgc;

    imgv[*imgc] = NULL;

    return imgv;
    }

static char **add_dir(char **imgv, int *imgc, char const *dir, loop_check_t const *parent)
    {
    DIR *d;
    struct dirent *item;
    char *path;
    loop_check_t lc;

    d = opendir(dir);
    if(NULL == d)
        {
        perror(dir);
        return imgv;
        }

    while((item = readdir(d)) != NULL)
        {
        if('.' == item->d_name[0])
            continue;

        path = e_malloc(strlen(dir) + 1 + strlen(item->d_name) + 1);

        strcat(strcat(strcpy(path, dir), "/"), item->d_name);

        if(is_dir(path, &lc.st))
            {
            lc.parent = parent;

            if(!is_loop(&lc))
                imgv = add_dir(imgv, imgc, path, &lc);
            }
        else if(is_image(path))
            imgv = add_file(imgv, imgc, path);

        free(path);
        }

    closedir(d);

    return imgv;
    }

static int is_dir(char const *path, struct stat *st)
    {
    return stat(path, st) == 0 && S_ISDIR(st->st_mode);
    }

static int is_loop(loop_check_t const *lc)
    {
    dev_t dev;
    ino_t ino;

    dev = lc->st.st_dev;
    ino = lc->st.st_ino;

    lc = lc->parent;

    while(lc != NULL)
        {
        if(lc->st.st_dev == dev && lc->st.st_ino == ino)
            return 1;

        lc = lc->parent;
        }

    return 0;
    }

static int is_image(char const *path)
    {
    char const *ext[] = 
        {
        ".bmp",
        ".gif",
        ".jpg",
        ".jpeg",
        ".pcx",
        ".png",
        ".pnm",
        ".tga",
        ".tif",
        ".tiff",
        ".xcf",
        ".xpm",
        ".webp"
        };
    int path_len, ext_len;
    unsigned i;

    path_len = strlen(path);

    for(i = 0; i < sizeof(ext) / sizeof(*ext); ++i)
        {
        ext_len = strlen(ext[i]);
        if(path_len >= ext_len && 0 == strcasecmp(path + path_len - ext_len, ext[i]))
            return 1;
        }

    return 0;
    }

void image_list_sort(char **imgv, int imgc)
    {
    qsort(imgv, imgc, sizeof(*imgv), cmp_files);
    }

static int cmp_files(const void *a, const void *b)
    {
    return strcoll(*(char **)a, *(char **)b);
    }

void image_list_shuffle(char **imgv, int imgc)
    {
    int i, r;
    char *tmp;

    srand((unsigned)time(NULL));

    for(i = 0; i < imgc - 1; ++i)
        {
        r = i + rand() % (imgc - i);
        tmp = imgv[i];
        imgv[i] = imgv[r];
        imgv[r] = tmp;
        }
    }

void image_dirs_shuffle(char **imgv, int imgc)
    {
    char **dirs = NULL;
    int dirc = 0;
    int img_idx;
    char **imgv_tmp;
    int dir_idx;
    int tmp_idx;

    /* Make list of directories */
    for(img_idx = 0; img_idx < imgc; ++img_idx)
        {
        if(!dir_present(dirs, dirc, imgv[img_idx]))
            {
            dirs = e_realloc(dirs, sizeof(*dirs) * (dirc + 1));
            dirs[dirc++] = dir_name(imgv[img_idx]);
            }
        }

    /* Shuffle directories */
    image_list_shuffle(dirs, dirc);

    /* Allocate new list of files */
    imgv_tmp = e_malloc(sizeof(*imgv_tmp) * imgc);
    tmp_idx = 0;

    /* For each directory */
    for(dir_idx = 0; dir_idx < dirc; ++dir_idx)
        {
        /* Add files from this directory to new list */
        for(img_idx = 0; img_idx < imgc; ++img_idx)
            {
            if(dir_match(dirs[dir_idx], imgv[img_idx]))
                imgv_tmp[tmp_idx++] = imgv[img_idx];
            }
        }

    /* Replace old file list with new one */
    memcpy(imgv, imgv_tmp, imgc * sizeof(*imgv));

    free(imgv_tmp);

    for(dir_idx = 0; dir_idx < dirc; ++dir_idx)
        free(dirs[dir_idx]);

    free(dirs);
    }

static char *dir_name(char const *path)
    {
    char *dir;
    char *e = strrchr(path, '/');

    if(NULL == e)
        return e_strdup("");

    dir = e_malloc(e - path + 2);
    memcpy(dir, path, e - path + 1);
    dir[e - path + 1] = 0;

    return dir;
    }

static int dir_present(char **dirs, int dirc, char const *path)
    {
    int i;

    if(NULL == dirs || 0 == dirc)
        return 0;

    for(i = dirc - 1; i >= 0; --i)
        {
        if(dir_match(dirs[i], path))
            return 1;
        }

    return 0;
    }

static int dir_match(char const *dir, char const *path)
    {
    size_t len = strlen(dir);

    if(len != 0 && memcmp(dir, path, len) != 0)
        return 0;

    if(strchr(path + len, '/') != NULL)
        return 0;

    return 1;
    }

SDL_Surface **image_frames_alloc(int n)
    {
    SDL_Surface **framev;

    framev = e_malloc(n * sizeof(*framev));

    memset(framev, 0, n * sizeof(*framev));

    return framev;
    }

void image_frames_free(SDL_Surface **framev, int n)
    {
    if(framev != NULL)
        {
        for(--n; n >= 0; --n)
            {
            if(framev[n] != NULL)
                SDL_FreeSurface(framev[n]);
            }

        free(framev);
        }
    }

void image_random(SDL_Surface *img)
    {
    int x, y;
    Uint32 *pdst;

    pdst = img->pixels;

    for(y = 0; y < img->h; ++y)
        {
        for(x = 0; x < img->w; ++x)
            pdst[y * img->pitch / 4 + x] =
                SDL_MapRGBA(img->format, rand() % 256, rand() % 256, rand() % 256, SDL_ALPHA_OPAQUE);
        }
    }

void image_black(SDL_Surface *img)
    {
    int x, y;
    Uint32 *pdst;

    pdst = img->pixels;

    for(y = 0; y < img->h; ++y)
        {
        for(x = 0; x < img->w; ++x)
            pdst[y * img->pitch / 4 + x] = SDL_MapRGBA(img->format, 0, 0, 0, SDL_ALPHA_OPAQUE);
        }
    }

int image_cmp(SDL_Surface *a, SDL_Surface *b)
    {
    int x, y;
    Uint32 *pa, *pb;

    if(a->w != b->w || a->h != b->h ||
        a->format->BytesPerPixel != 4 || b->format->BytesPerPixel != 4)
        return 1;

    pa = a->pixels;
    pb = b->pixels;

    for(y = 0; y < a->h; ++y)
        {
        for(x = 0; x < a->w; ++x)
            {
            if((pa[y * a->pitch / 4 + x] & ~a->format->Amask)
                != (pb[y * b->pitch / 4 + x] & ~b->format->Amask))
                return 1;
            }
        }

    return 0;
    }

int image_format_ok(SDL_PixelFormat *a, SDL_PixelFormat *b)
    {
    return a->Rmask == b->Rmask
        && a->Gmask == b->Gmask
        && a->Bmask == b->Bmask
        && a->Amask == b->Amask;
    }

void image_blur(SDL_Surface *img, frei0r_plugin_t *blur)
    {
    double strength = 0.03;
    f0r_instance_t inst = NULL;
    SDL_Surface *dst = NULL;
    SDL_Surface *src = NULL;
    SDL_Surface *tmp = NULL;

    if(blur != NULL)
        inst = blur->f_construct(img->w, img->h);

    if(NULL == inst)
        return;

    if(F0R_COLOR_MODEL_BGRA8888 == blur->info.color_model)
        dst = image_alloc_bgra(img->w, img->h);
    else
        dst = image_alloc_rgba(img->w, img->h);

    if(image_format_ok(img->format, dst->format))
        src = img;
    else
        {
        tmp = image_convert(img, dst->format);
        src = tmp;
        }

    blur->f_set_param_value(inst, &strength, 0);
    blur->f_update(inst, 0.0, src->pixels, dst->pixels);
    blur->f_destruct(inst);

    if(tmp != NULL)
        {
        SDL_FreeSurface(tmp);
        tmp = image_convert(dst, img->format);
        SDL_FreeSurface(dst);
        dst = tmp;
        }

    memcpy(img->pixels, dst->pixels, img->w * img->h * img->format->BytesPerPixel);

    SDL_FreeSurface(dst);
    }

#define FILL_STEP 32

int image_fill(SDL_Surface *dst, SDL_Surface *src, frei0r_plugin_t *blur)
    {
    int border;
    int part;

    border = image_fit(dst, src, 1, 1, 1);

    if(-1 == border)
        return -1;

    for(part = border - FILL_STEP; part > 0; part -= FILL_STEP)
        image_fit(dst, src, part, border, 1);

    image_blur(dst, blur);

    return image_fit(dst, src, 0, 1, 0);
    }

int image_fit(SDL_Surface *dst, SDL_Surface *src, int om, int od, int fast)
    {
    int crop_w, crop_h;
    int km, kd;
    int w, h;
    int dx, dy;
    int ret;

    if(src->format->BytesPerPixel != 4)
        {
        fprintf(stderr, "image_fit: bad src format = %d\n", src->format->BytesPerPixel);
        return -1;
        }

    if(dst->format->BytesPerPixel != 4)
        {
        fprintf(stderr, "image_fit: bad dst format = %d\n", dst->format->BytesPerPixel);
        return -1;
        }

    crop_w = dst->clip_rect.w;
    crop_h = dst->clip_rect.h;

    if(crop_w > dst->w || om != 0)
        crop_w = dst->w;
    if(crop_h > dst->h || om != 0)
        crop_h = dst->h;


    if(crop_w * src->h < crop_h * src->w)
        {
        km = crop_w;
        kd = src->w;
        }
    else
        {
        km = crop_h;
        kd = src->h;
        }

    w = (src->w * km + kd / 2) / kd;
    h = (src->h * km + kd / 2) / kd;

    dx = (crop_w - w) / 2;
    dy = (crop_h - h) / 2;

    ret = dx | dy;

    if(0 == om || 0 == od)
        {
        if(!fast && km < kd)
            image_downscale(dst, src, w, h, dx, dy);
        else if(!fast && km > kd && src->w > 3 && src->h > 3)
            image_upscale(dst, src, w, h, dx, dy);
        else
            image_fastscale(dst, src, w, h, dx, dy, km, kd, 0, 0);
        }
    else
        image_overscale(dst, src, w, h, dx, dy, om, od);

    if(0 == od && 0 == om && dy > 0)
        {
        Uint32 *pdst = dst->pixels;
        int x, y;

        for(y = 0; y < dy; ++y)
            {
            for(x = 0; x < crop_w; ++x)
                    *pdst++ = dst->format->Amask;

            pdst += dst->pitch / 4 - crop_w;
            }

        pdst += h * dst->pitch / 4;

        for(y = dy + h; y < crop_h; ++y)
            {
            for(x = 0; x < crop_w; ++x)
                    *pdst++ = dst->format->Amask;

            pdst += dst->pitch / 4 - crop_w;
            }
        }
    else if(0 == od && 0 == om && dx > 0)
        {
        Uint32 *pdst = dst->pixels;
        int x, y;

        for(y = 0; y < crop_h; ++y)
            {
            for(x = 0; x < dx; ++x)
                    *pdst++ = dst->format->Amask;

            pdst += w;

            for(x = dx + w; x < crop_w; ++x)
                    *pdst++ = dst->format->Amask;

            pdst += dst->pitch / 4 - crop_w;
            }
        }

    return ret;
    }

static void image_overscale(SDL_Surface *dst, SDL_Surface *src, int w, int h, int dx, int dy, int om, int od)
    {
    int km, kd;
    int sx, sy;

    if(dy > 0)
        {
        h += (dy * 2 * om + od / 2) / od;

        km = h;
        kd = src->h;

        dx = 0;
        dy = (dy * (od - om) + od / 2) / od;

        sx = (src->w - (w * kd + km / 2) / km + 1) / 2;
        sy = 0;
        }
    else
        {
        w += (dx * 2 * om + od / 2) / od;

        km = w;
        kd = src->w;

        dx = (dx * (od - om) + od / 2) / od;
        dy = 0;

        sx = 0;
        sy = (src->h - (h * kd + km / 2) / km + 1) / 2;
        }

    image_fastscale(dst, src, w, h, dx, dy, km, kd, sx, sy);
    }

static void image_fastscale(
    SDL_Surface *dst, SDL_Surface *src, int w, int h, int dx, int dy, int km, int kd, int sx, int sy)
    {
    Uint32 *psrc = src->pixels;
    Uint32 *pdst = dst->pixels;
    int x, y;

    psrc += sy * src->pitch / 4 + sx;
    pdst += dy * dst->pitch / 4 + dx;

    for(y = 0; y < h; ++y)
        {
        Uint32 *ysrc = psrc + y * kd / km * src->pitch / 4;

        for(x = 0; x < w; ++x)
            *pdst++ = ysrc[x * kd / km];

        pdst += dst->pitch / 4 - w;
        }
    }

static void image_upscale(SDL_Surface *dst, SDL_Surface *src, int w, int h, int dx, int dy)
    {
    int km, kd;
    Uint32 *psrc, *pdst;
    Uint32 *ep;
    int dh;

    km = w * 2;
    kd = src->w * 2;

    dh = dy + h - src->h;

    psrc = src->pixels;
    ep = psrc + src->pitch / 4 * src->h;

    pdst = dst->pixels;
    pdst += dh * dst->pitch / 4 + dx;

    while(psrc < ep)
        {
        int a1[3 * 4];
        int a2[3 * 4];
        int px0, px1, px2;
        int offset = (kd - km * 3) / 2;
        Uint32 *ers = psrc + src->w;
        Uint32 *erd = pdst + w;

        px0 = *psrc++;
        px1 = *psrc++;
        px2 = *psrc++;

        par32(a2, px0, px1, px2);
        px0 = px1;
        px1 = px2;

        while(offset < 0)
            {
            *pdst++ = get_pixel1(a2, offset, km);
            offset += kd;
            }

        while(psrc < ers)
            {
            px2 = *psrc++;

            memcpy(a1, a2, sizeof(a1));

            par32(a2, px0, px1, px2);
            px0 = px1;
            px1 = px2;

            while(offset < km)
                {
                *pdst++ = get_pixel2(a1, a2, offset, km);
                offset += kd;
                }

            offset -= km;
            }

        while(pdst < erd)
            {
            *pdst++ = get_pixel1(a2, offset, km);
            offset += kd;
            }

        psrc += src->pitch / 4 - src->w;
        pdst += dst->pitch / 4 - w;
        }

    km = h * 2;
    kd = src->h * 2;

    psrc = dst->pixels;
    psrc += dh * dst->pitch / 4 + dx;

    dh = dst->pitch / 4;

    pdst = dst->pixels;
    pdst += dy * dh + dx;
    ep = pdst + w;

    while(pdst < ep)
        {
        int a1[3 * 4];
        int a2[3 * 4];
        int px0, px1, px2;
        int offset = (kd - km * 3) / 2;
        Uint32 *ecs = psrc + src->h * dh;
        Uint32 *ecd = pdst + h * dh;;

        px0 = *psrc;
        psrc += dh;
        px1 = *psrc;
        psrc += dh;
        px2 = *psrc;
        psrc += dh;

        par32(a2, px0, px1, px2);
        px0 = px1;
        px1 = px2;

        while(offset < 0)
            {
            *pdst = get_pixel1(a2, offset, km);
            pdst += dh;
            offset += kd;
            }

        while(psrc < ecs)
            {
            px2 = *psrc;
            psrc += dh;

            memcpy(a1, a2, sizeof(a1));

            par32(a2, px0, px1, px2);
            px0 = px1;
            px1 = px2;

            while(offset < km)
                {
                *pdst = get_pixel2(a1, a2, offset, km);
                pdst += dh;
                offset += kd;
                }

            offset -= km;
            }

        while(pdst < ecd)
            {
            *pdst = get_pixel1(a2, offset, km);
            pdst += dh;
            offset += kd;
            }

        psrc -= dh * src->h - 1;
        pdst -= dh * h - 1;
        }
    }

static inline void par(int a[3], int y1, int y2, int y3)
    {
    int k;

    k = (y3 + y1 + 1) / 2;
    a[0] = k - y2;
    a[1] = k - y1;
    a[2] = y2;
    }

static inline void par32(int a[3 * 4], Uint32 px0, Uint32 px1, Uint32 px2)
    {
    par(a, px0 & 0xFF, px1 & 0xFF, px2 & 0xFF);
    par(a + 3, (px0 >> 8) & 0xFF, (px1 >> 8) & 0xFF, (px2 >> 8) & 0xFF);
    par(a + 6, (px0 >> 16) & 0xFF, (px1 >> 16) & 0xFF, (px2 >> 16) & 0xFF);
    par(a + 9, (px0 >> 24) & 0xFF, (px1 >> 24) & 0xFF, (px2 >> 24) & 0xFF);
    }

static inline Uint32 get_pixel1(int const a[3 * 4], int dx, int div)
    {
    return get_color1(a, dx, div) | 
        (get_color1(a + 3, dx, div) << 8) |
        (get_color1(a + 6, dx, div) << 16) |
        (get_color1(a + 9, dx, div) << 24);
    }

static inline Uint32 get_pixel2(int const a1[3 * 4], int const a2[3 * 4], int dx, int div)
    {
    return get_color2(a1, a2, dx, div) | 
        (get_color2(a1 + 3, a2 + 3, dx, div) << 8) |
        (get_color2(a1 + 6, a2 + 6, dx, div) << 16) |
        (get_color2(a1 + 9, a2 + 9, dx, div) << 24);
    }

static inline Uint32 get_color1(int const a[3], int dx, int div)
    {
    int d2;
    int c;

    d2 = div / 2;

    c = (((a[0] * dx + d2) / div + a[1]) * dx + d2) / div + a[2];

    if(c < 0)
        c = 0;
    else if(c > 255)
        c = 255;

    return (Uint32)c;
    }

static inline Uint32 get_color2(int const a1[3], int const a2[3], int dx, int div)
    {
    int ndx;
    int d2;
    int c1, c2;

    ndx = dx - div;
    d2 = div / 2;

    c1 = (((a1[0] * dx + d2) / div + a1[1]) * dx * ndx + d2) / div + a1[2] * ndx;
    c2 = (((a2[0] * ndx + d2) / div + a2[1]) * ndx * dx + d2) / div + a2[2] * dx;

    c2 = (c2 - c1 + d2) / div;

    if(c2 < 0)
        c2 = 0;
    else if(c2 > 255)
        c2 = 255;

    return (Uint32)c2;
    }

static void image_downscale(SDL_Surface *dst, SDL_Surface *src, int w, int h, int dx, int dy)
    {
    int kdx, kmx;
    int kdy, kmy;
    Uint32 *pdst, *rsrc;
    Uint32 *ep;
    int *rsum, *esum;
    int ly, ky;

    kmx = w;
    kdx = src->w;

    kmy = h;
    kdy = src->h;

    rsum = e_malloc(4 * w * sizeof(*rsum));
    memset(rsum, 0, 4 * w * sizeof(*rsum));

    esum = rsum + w * 4;

    rsrc = src->pixels;

    pdst = dst->pixels;
    pdst += dst->pitch / 4 * dy + dx;

    ep = pdst + dst->pitch / 4 * h;

    ly = kdy;
    ky = kmy;

    while(pdst < ep)
        {
        int lx = 0;
        Uint32 *psrc = rsrc;
        int *sum;

        for(sum = rsum; sum < esum; sum += 4)
            {
            add32(sum, *psrc, (kmx - lx) * ky, kdx);
            ++psrc;

            lx += kdx - kmx;

            while(lx > kmx)
                {
                add32(sum, *psrc, kmx * ky, kdx);
                ++psrc;

                lx -= kmx;
                }

            add32(sum, *psrc, lx * ky, kdx);
            }

        ly -= ky;

        if(0 == ly)
            {
            for(sum = rsum; sum < esum; sum += 4)
                *pdst++ = div32(sum, kdy);

            pdst += dst->pitch / 4 - w;

            memset(rsum, 0, 4 * w * sizeof(*rsum));

            ky = kmy - ky;
            ly = kdy;
            }
        else
            {
            if(ly < kmy)
                ky = ly;
            else
                ky = kmy;

            rsrc += src->pitch / 4;
            }
        }

    free(rsum);
    }

static inline void add32(int sum[4], Uint32 px, int mul, int div)
    {
    sum[0] += ((px & 0xFF) * mul + div / 2) / div;
    sum[1] += (((px >> 8) & 0xFF) * mul + div / 2) / div;
    sum[2] += (((px >> 16) & 0xFF) * mul + div / 2) / div;
    sum[3] += (((px >> 24) & 0xFF) * mul + div / 2) / div;
    }

static inline Uint32 div32(int const sum[4], int div)
    {
    return ((sum[0] + div / 2) / div)
        | (((sum[1] + div / 2) / div) << 8)
        | (((sum[2] + div / 2) / div) << 16)
        | (((sum[3] + div / 2) / div) << 24);
    }

SDL_Surface *load_resource(void const *data, int size)
    {
    SDL_RWops *rw;
    SDL_Surface *img;

    rw = SDL_RWFromConstMem(data, size);
    if(NULL == rw)
        {
        fprintf(stderr, "SDL_RWFromConstMem failed: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
        }

    img = IMG_Load_RW(rw, 1);
    if(NULL == img)
        {
        fprintf(stderr, "IMG_LoadRW failed: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
        }

    return img;
    }
