/*
This file is a part of sleid0r -- slideshow viewer/generator.
Copyright (C) 2013-2015 Vadim Druzhin <cdslow@mail.ru>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include <SDL_image.h>
#include <frei0r.h>
#include <unistd.h>
#include <locale.h>
#include <time.h>
#include "err.h"
#include "images.h"
#include "mod_frei0r.h"
#include "options.h"
#include "slide.h"
#include "resource.h"

static char const *Tag = "sleid0r";

typedef struct state_s
    {
    /* Args */
    char **imgv;
    int imgc;
    int imgi;
    int img_failed;
    int shuffle;
    int d_shuffle;
    int repeat;
    int e_shuffle;
    /* Images */
    SDL_Surface *screen;
    Uint32 fs;
    char *fs_id;
    int fs_all;
    int fs_delay;
    int win_w;
    int win_h;
    int display_w;
    int display_h;
    int all_w;
    int all_h;
    SDL_Surface *img1;
    SDL_Surface *img2;
    int load_w;
    int load_h;
    SDL_Thread *loader;
    /* Effects */
    frei0r_plugin_t *blend;
    frei0r_plugin_t *blur;
    frei0r_plugin_t **mixv;
    int mixi;
    /* Frames */
    SDL_Surface **framev;
    int framec;
    int framei;
    SDL_Thread *renderer;
    int stop;
    Uint32 duration;
    Uint32 ticks;
    Uint32 show_ticks;
    Uint32 ms_frame;
    SDL_TimerID timer_frame;
    /* Timer */
    Uint32 ms_img;
    SDL_TimerID timer_img;
    /* Control */
    int space_is_pause;
    int pause;
    int next;
    int back;
    Uint32 pause_ticks;
    /* Icons */
    SDL_Surface *icon_pause;
    } state_t;

enum EVENT_CODE
    {
    EC_INIT = 1,
    EC_IMG_LOAD,
    EC_IMG_READY,
    EC_IMG_FAILED,
    EC_RENDER,
    EC_FRAMES_READY,
    EC_DISPLAY,
    EC_NEXT_FRAME,
    EC_SHOW_FRAME,
    EC_IMG_DELAY,
    EC_CONTINUE,
    EC_FULLSCREEN,
    EC_PAUSE,
    EC_NEXT_OR_BACK
    };

static void list_mixers(void);
static int show_slides(state_t *state);
static int write_slides(param_t *params, state_t *state);
static int write_slides_2(char *path, char *num, int num_w, int count, state_t *state);
static int on_videoresize(SDL_ResizeEvent *resize, state_t *state);
static int on_keyup(SDL_KeyboardEvent *key, state_t *state);
static int on_userevent(SDL_UserEvent *user);
static int update_image(state_t *state);
static int show_image(SDL_Surface *screen, SDL_Surface *image, int fast, SDL_Surface *icon);
static Uint32 delay_image(Uint32 interval, void *data);
static Uint32 delay_frame(Uint32 interval, void *data);
static void stop_timers(state_t *state);
static int do_init(state_t *state);
static int do_load(state_t *state);
static int do_ready(state_t *state);
static int do_failed(state_t *state);
static int do_render(state_t *state);
static int do_frames_ready(state_t *state);
static int do_continue(state_t *state);
static int do_display(state_t *state);
static int do_next_frame(state_t *state);
static int do_show_frame(state_t *state);
static int do_delay(state_t *state);
static int do_fullscreen(state_t *state);
static int do_pause(state_t *state);
static int do_next_or_back(state_t *state);
static int load_image(void *data);
static int load_img2(state_t *state);
static int render_frames(void *data);
static frei0r_plugin_t **select_mixers_by_test(frei0r_list_t *fl);
static frei0r_plugin_t **select_mixers_by_list(frei0r_list_t *fl, char **list);
static void shuffle_mixers(frei0r_plugin_t **sel);
static void progress_tty(unsigned i, unsigned n);

int main(int argc, char *argv[])
    {
    param_t params;
    state_t state;
    frei0r_list_t *plugins;
    int ret;

    (void)argc; /* Unused */

    setlocale(LC_ALL, "");

    init_params(&params);

    if(!read_config(&params, "sleid0r.cfg"))
        exit(EXIT_FAILURE);

    if(!parse_options(&params, argv + 1))
        exit(EXIT_FAILURE);

    if(params.help)
        {
        usage();
        return EXIT_SUCCESS;
        }

    if(params.list)
        {
        list_mixers();
        return EXIT_SUCCESS;
        }

    memset(&state, 0, sizeof(state));

    state.icon_pause = load_resource(ImagePause, ImagePauseSize);

    state.imgv = image_list(&state.imgc, argv + 1);

    if(params.playlist)
        state.imgv = image_list_append(state.imgv, &state.imgc, params.playlist);

    state.imgi = -1;

    if(0 == state.imgc)
        {
        fprintf(stderr, "No images!\n");
        exit(EXIT_FAILURE);
        }

    if(params.sort)
        image_list_sort(state.imgv, state.imgc);
    else if(params.shuffle)
        state.shuffle = 1;

    if(params.d_shuffle)
        state.d_shuffle = 1;

    if(params.repeat)
        state.repeat = 1;

    if(params.e_shuffle)
        state.e_shuffle = 1;

    state.ms_img = (int)(params.delay * 1000.0 + 0.5);

    plugins = frei0r_get_plugins();

    if(0 == plugins->mixer_count)
        fprintf(stderr, "No transitions!\n");
    else
        {
        state.blend = frei0r_search(plugins, "blend");
        if(NULL == state.blend)
            fprintf(stderr, "'blend' transition not found\n");

        if(NULL == params.selection)
            state.mixv = select_mixers_by_test(plugins);
        else
            state.mixv = select_mixers_by_list(plugins, params.selection);

        if(NULL == state.mixv[0])
            fprintf(stderr, "No transitions!\n");
        }

    if(!params.black)
        {
        if(plugins->filter_count > 0)
            state.blur = frei0r_search(plugins, "Squareblur");

        if(NULL == state.blur)
            fprintf(stderr, "'Squareblur' filter not found\n");
        }

    state.mixi = -1;

    state.duration = (int)(params.duration * 1000.0 + 0.5);
    state.framec = (int)(params.duration * params.fps + 0.5);
    state.framei = state.framec;
    if(state.framec > 0)
        state.framev = image_frames_alloc(state.framec);

    state.win_w = params.w;
    state.win_h = params.h;

    if(params.fs)
        {
        state.fs = SDL_FULLSCREEN;

        if(params.fs > 1)
            state.fs_all = 1;
        }
    else
        state.fs = SDL_RESIZABLE;

    state.fs_id = params.fs_id;

    state.space_is_pause = params.space_is_pause;

    if(NULL == params.prefix)
        ret = show_slides(&state);
    else
        ret = write_slides(&params, &state);

    state.stop = 1;

    if(state.timer_frame != NULL)
        SDL_RemoveTimer(state.timer_frame);

    if(state.timer_img != NULL)
        SDL_RemoveTimer(state.timer_img);

    if(state.loader != NULL)
        SDL_WaitThread(state.loader, NULL);

    if(state.renderer != NULL)
        SDL_WaitThread(state.renderer, NULL);

    if(state.img1 != NULL)
        SDL_FreeSurface(state.img1);

    if(state.img2 != NULL)
        SDL_FreeSurface(state.img2);

    if(state.mixv != NULL)
        free(state.mixv);

    image_frames_free(state.framev, state.framec);

    frei0r_free_plugins(plugins);

    image_list_free(state.imgv, state.imgc);

    SDL_FreeSurface(state.icon_pause);

    free_params(&params);

    return ret < 0 ? EXIT_FAILURE : EXIT_SUCCESS;
    }

static int show_slides(state_t *state)
    {
    SDL_VideoInfo const *vi;
    SDL_Rect **modes;
    SDL_Event event;

    setenv("SDL_VIDEO_FULLSCREEN_DISPLAY", state->fs_id, 0);
    setenv("SDL_VIDEO_X11_XRANDR", "0", 0);

    if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0)
        {
        fprintf(stderr, "SDL init failed: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
        }

    atexit(SDL_Quit);

    vi = SDL_GetVideoInfo();
    state->all_w = vi->current_w;
    state->all_h = vi->current_h;
    state->display_w = vi->current_w;
    state->display_h = vi->current_h;

    modes = SDL_ListModes(NULL, SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_FULLSCREEN);
    if(modes != NULL && modes != (SDL_Rect **)-1)
        {
        while(*modes)
            {
            state->display_w = (*modes)->w;
            state->display_h = (*modes)->h;
            ++modes;
            }
        }

    event.user.type = SDL_USEREVENT;
    event.user.code = EC_INIT;
    event.user.data1 = state;
    event.user.data2 = NULL;

    do
        {
        if(SDL_QUIT == event.type)
            return 0;
        else if(SDL_USEREVENT == event.type)
            {
#if 0
            if(event.user.code == EC_NEXT_FRAME)
                fprintf(stderr, "[%u] EC_NEXT_FRAME: framei=%u ticks=%u\n", SDL_GetTicks(), state->framei, state->ticks);
            else if(event.user.code == EC_SHOW_FRAME)
                fprintf(stderr, "[%u] EC_SHOW_FRAME: framei=%u ticks=%u\n", SDL_GetTicks(), state->framei, state->ticks);
            else if(event.user.code == EC_DISPLAY)
                {
                fprintf(stderr, "[%u] EC_DISPLAY: image=%u/%u frame=%u/%u next=%u\n", SDL_GetTicks(),
                    state->imgi, state->imgc, state->framei, state->framec, state->next);
                if(state->framei < state->framec)
                    fprintf(stderr, "  TRANSITION: %s\n", state->mixv[state->mixi]->info.name);
                else
                    fprintf(stderr, "  IMAGE: %s\n", state->imgv[state->imgi]);
                }
            else if(event.user.code == EC_RENDER)
                {
                fprintf(stderr, "[%u] EC_RENDER: renderer=%p\n", SDL_GetTicks(), (void *)state->renderer);
                }
            else
                fprintf(stderr, "[%u] USERVENT: code=%0d\n", SDL_GetTicks(), event.user.code);
#endif
            if(on_userevent(&event.user) < 0)
                break;
            }
        else if(SDL_VIDEORESIZE == event.type)
            {
            if(on_videoresize(&event.resize, state) < 0)
                break;
            }
        else if(SDL_KEYUP == event.type)
            {
            if(on_keyup(&event.key, state) < 0)
                break;
            }
        }
    while(SDL_WaitEvent(&event) == 1);

    return -1;
    }

static int write_slides(param_t *params, state_t *state)
    {
    char *path = NULL, *num = NULL;
    int num_w = 0;
    int count = (int)(params->delay * params->fps + 0.5);
    int i;
    int ret = 0;

    if(state->shuffle)
        image_list_shuffle(state->imgv, state->imgc);

    if(state->d_shuffle)
        image_dirs_shuffle(state->imgv, state->imgc);

    state->load_w = params->w;
    state->load_h = params->h;

    state->img2 = image_alloc_rgba(state->load_w, state->load_h);

    do
        {
        ++state->imgi;
        progress_tty(state->imgi, state->imgc);
        }
    while(state->imgi < state->imgc && load_img2(state) < 0);

    if(state->imgi == state->imgc)
        {
        fprintf(stderr, "No images!\n");
        return -1;
        }

    ret = -state->imgi;

    if(params->prefix[0] != '-' || params->prefix[1] != 0)
        {
        for(i = state->imgc * (count + state->framec); i != 0; i /= 10)
            ++num_w;

        path = prepare_path(params->prefix, num_w);

        num = path + strlen(params->prefix);
        }

    ret += write_slides_2(path, num, num_w, count, state);

    if(path != NULL)
        free(path);

    return ret;
    }

static int write_slides_2(char *path, char *num, int num_w, int count, state_t *state)
    {
    unsigned idx = 1;
    int ret = 0;
    SDL_Surface *tmp;
    int i;

    if(0 == write_images(path, num, num_w, idx, state->img2, count))
        return -1;

    idx += count;

    state->img1 = state->img2;
    state->img2 = image_alloc_rgba(state->load_w, state->load_h);

    for(i = 0; i < state->framec; ++i)
        state->framev[i] = image_alloc_rgba(state->load_w, state->load_h);

    while(++state->imgi < state->imgc)
        {
        progress_tty(state->imgi, state->imgc);

        if(load_img2(state) < 0)
            {
            ret = -1;
            continue;
            }

        ++state->mixi;
        if(NULL == state->mixv[state->mixi])
            state->mixi = 0;

        if(0 == state->mixi && state->e_shuffle)
            shuffle_mixers(state->mixv);
        
        if(state->framec > 0 && state->mixv[state->mixi] != NULL)
            {
            make_frames(state->framev, state->framec, state->img1, state->img2,
                state->mixv[state->mixi], state->blend, &state->stop);

            if(0 == write_frames(path, num, num_w, idx, state->framev, state->framec))
                return -1;

            idx += state->framec;
            }

        if(0 == write_images(path, num, num_w, idx, state->img2, count))
            return -1;

        idx += count;

        tmp = state->img1;
        state->img1 = state->img2;
        state->img2 = tmp;
        }

    progress_tty(state->imgi, state->imgc);

    return ret;
    }

static int on_videoresize(SDL_ResizeEvent *resize, state_t *state)
    {
    state->screen = SDL_SetVideoMode(resize->w, resize->h, 32,
        SDL_HWSURFACE | SDL_DOUBLEBUF | state->fs);
    if(NULL == state->screen)
        {
        fprintf(stderr, "SetVideoMode failed: %s\n", SDL_GetError());
        return -1;
        }

    update_image(state);

    if(SDL_FULLSCREEN == state->fs)
        SDL_ShowCursor(SDL_DISABLE);
    else
        SDL_ShowCursor(SDL_ENABLE);

    return 0;
    }

static int on_keyup(SDL_KeyboardEvent *key, state_t *state)
    {
    SDL_Event event;

    event.user.type = SDL_USEREVENT;
    event.user.data1 = state;
    event.user.data2 = NULL;

    switch(key->keysym.sym)
        {
    case SDLK_ESCAPE:
    case SDLK_q:
        event.type = SDL_QUIT;
        break;

    case SDLK_f:
        if(key->keysym.mod & KMOD_SHIFT)
            state->fs_all = 1;
        else
            state->fs_all = 0;

        event.user.code = EC_FULLSCREEN;
        break;

    case SDLK_PAUSE: 
        state->pause = !state->pause;
        event.user.code = EC_PAUSE;
        break;

    case SDLK_p:
        if(state->pause)
            return 0;

        state->pause = 1;
        event.user.code = EC_PAUSE;
        break;

    case SDLK_c:
        if(!state->pause)
            return 0;

        state->pause = 0;
        event.user.code = EC_PAUSE;
        break;

    case SDLK_PAGEDOWN:
    case SDLK_n:
        state->next = 1;
        state->back = 0;
        event.user.code = EC_NEXT_OR_BACK;
        break;

    case SDLK_PAGEUP:
    case SDLK_b:
    case SDLK_BACKSPACE:
        state->next = 1;
        state->back = 1;
        event.user.code = EC_NEXT_OR_BACK;
        break;

    case SDLK_SPACE:
        if(state->space_is_pause)
            {
            state->pause = !state->pause;
            event.user.code = EC_PAUSE;
            }
        else
            {
            state->next = 1;
            state->back = key->keysym.mod & KMOD_SHIFT ? 1 : 0;
            event.user.code = EC_NEXT_OR_BACK;
            }
        break;

    default:
        return 0;
        }

    return SDL_PushEvent(&event);
    }

static int on_userevent(SDL_UserEvent *user)
    {
    switch(user->code)
        {
    case EC_INIT:
        return do_init(user->data1);

    case EC_IMG_LOAD:
        return do_load(user->data1);

    case EC_IMG_READY:
        return do_ready(user->data1);

    case EC_IMG_FAILED:
        return do_failed(user->data1);

    case EC_RENDER:
        return do_render(user->data1);

    case EC_FRAMES_READY:
        return do_frames_ready(user->data1);

    case EC_DISPLAY:
        return do_display(user->data1);

    case EC_NEXT_FRAME:
        return do_next_frame(user->data1);

    case EC_SHOW_FRAME:
        return do_show_frame(user->data1);

    case EC_IMG_DELAY:
        return do_delay(user->data1);

    case EC_CONTINUE:
        return do_continue(user->data1);

    case EC_FULLSCREEN:
        return do_fullscreen(user->data1);

    case EC_PAUSE:
        return do_pause(user->data1);

    case EC_NEXT_OR_BACK:
        return do_next_or_back(user->data1);

    default:
        fprintf(stderr, "Unknown user event code: %d\n", user->code);
        return -1;
        }

    return 0;
    }

static int do_init(state_t *state)
    {
    SDL_Event event;

    event.resize.type = SDL_VIDEORESIZE;
    event.resize.w = state->win_w;
    event.resize.h = state->win_h;
    if(SDL_PushEvent(&event) < 0)
        return -1;

    if(SDL_FULLSCREEN == state->fs)
        {
        state->fs = SDL_RESIZABLE;

        event.user.type = SDL_USEREVENT;
        event.user.code = EC_FULLSCREEN;
        event.user.data1 = state;
        event.user.data2 = NULL;

        state->fs_delay = 10;

        if(SDL_PushEvent(&event) < 0)
            return -1;
        }

    event.user.type = SDL_USEREVENT;
    event.user.code = EC_IMG_LOAD;
    event.user.data1 = state;
    event.user.data2 = NULL;

    return SDL_PushEvent(&event);
    }

static int do_fullscreen(state_t *state)
    {
    SDL_Event event;

    if(SDL_RESIZABLE == state->fs)
        {
        state->fs = SDL_FULLSCREEN;
        state->win_w = state->screen->w;
        state->win_h = state->screen->h;

        event.resize.type = SDL_VIDEORESIZE;

        if(state->fs_all)
            {
            event.resize.w = state->all_w;
            event.resize.h = state->all_h;
            }
        else
            {
            event.resize.w = state->display_w;
            event.resize.h = state->display_h;
            }

        if(state->fs_delay)
            {
            SDL_Delay(state->fs_delay);
            state->fs_delay = 0;
            }
        }
    else
        {
        state->fs = SDL_RESIZABLE;

        event.resize.type = SDL_VIDEORESIZE;
        event.resize.w = state->win_w;
        event.resize.h = state->win_h;
        }

    return on_videoresize(&event.resize, state);
    }

static int do_load(state_t *state)
    {
    if(NULL == state->screen)
        {
        fprintf(stderr, "Screen not available!\n");
        return -1;
        }

    if(NULL != state->loader)
        SDL_WaitThread(state->loader, NULL);

    ++state->imgi;
    if(state->imgi == state->imgc)
        {
        if(!state->repeat)
            {
            state->stop = 1;
            return 0;
            }
        state->imgi = 0;
        }

    if(0 == state->imgi)
        {
        if(state->shuffle)
            image_list_shuffle(state->imgv, state->imgc);

        if(state->d_shuffle)
            image_dirs_shuffle(state->imgv, state->imgc);
        }

    state->load_w = state->screen->w;
    state->load_h = state->screen->h;
    
    state->loader = SDL_CreateThread(load_image, state);

    return 0;
    }

static int do_ready(state_t *state)
    {
    SDL_Event event;

    SDL_WaitThread(state->loader, NULL);
    state->loader = NULL;

    state->img_failed = 0;

    if(state->img1 != NULL && state->mixv[0] != NULL && state->framec > 0 && !state->next && !state->pause)
        {
        event.user.type = SDL_USEREVENT;
        event.user.code = EC_RENDER;
        event.user.data1 = state;
        event.user.data2 = NULL;

        return SDL_PushEvent(&event);
        }
    else if(!state->timer_img || state->next)
        {
        event.user.type = SDL_USEREVENT;
        event.user.code = EC_DISPLAY;
        event.user.data1 = state;
        event.user.data2 = NULL;

        return SDL_PushEvent(&event);
        }
    else
        return 0;
    }

static int do_failed(state_t *state)
    {
    SDL_Event event;

    ++state->img_failed;
    if(state->img_failed >= state->imgc)
        {
        fprintf(stderr, "No images!\n");
        return -1;
        }

    event.user.type = SDL_USEREVENT;
    event.user.code = EC_IMG_LOAD;
    event.user.data1 = state;
    event.user.data2 = NULL;

    return SDL_PushEvent(&event);
    }

static int do_render(state_t *state)
    {
    if(NULL == state->img1 || NULL == state->img2)
        {
        fprintf(stderr, "No images!\n");
        return -1;
        }

    if(NULL == state->mixv[0])
        {
        fprintf(stderr, "No transitions!\n");
        return -1;
        }

    if(NULL != state->renderer)
        SDL_WaitThread(state->renderer, NULL);

    ++state->mixi;
    if(NULL == state->mixv[state->mixi])
        state->mixi = 0;

    if(0 == state->mixi && state->e_shuffle)
        shuffle_mixers(state->mixv);
    
    state->framei = state->framec;

#if 0
    fprintf(stderr, "  RENDER: mixi=%d '%s'\n", state->mixi, state->mixv[state->mixi]->info.name);
#endif
    state->renderer = SDL_CreateThread(render_frames, state);

    return 0;
    }

static int do_frames_ready(state_t *state)
    {
    SDL_Event event;

    SDL_WaitThread(state->renderer, NULL);
    state->renderer = NULL;

    state->framei = 0;

    if(!state->timer_img || state->next)
        {
        event.user.type = SDL_USEREVENT;
        event.user.code = EC_DISPLAY;
        event.user.data1 = state;
        event.user.data2 = NULL;

        return SDL_PushEvent(&event);
        }
    else
        return 0;
    }

static int do_display(state_t *state)
    {
    SDL_Event event;
    SDL_Surface *tmp;

    if(state->pause && !state->next)
        {
        update_image(state);
        return 0;
        }

    if(state->next && (state->framei > 0 || state->back || state->pause))
        state->framei = state->framec;

    if(state->framei < state->framec)
        {
        SDL_WM_SetCaption(state->mixv[state->mixi]->info.name, Tag);

        state->next = 0;

        event.user.type = SDL_USEREVENT;
        event.user.code = EC_SHOW_FRAME;
        event.user.data1 = state;
        event.user.data2 = NULL;

        return SDL_PushEvent(&event);
        }
    else if(state->back)
        {
        state->back = 0;

        state->imgi -= 3;
        while(state->imgi < 0)
            state->imgi += state->imgc;

        event.user.type = SDL_USEREVENT;
        event.user.code = EC_IMG_LOAD;
        event.user.data1 = state;
        event.user.data2 = NULL;

        return SDL_PushEvent(&event);
        }
    else
        {
        SDL_WM_SetCaption(state->imgv[state->imgi], Tag);

        tmp = state->img1;
        state->img1 = state->img2;
        state->img2 = tmp;

        update_image(state);

        state->next = 0;

        event.user.type = SDL_USEREVENT;
        event.user.code = EC_IMG_LOAD;
        event.user.data1 = state;
        event.user.data2 = NULL;

        if(SDL_PushEvent(&event) < 0)
            return -1;

        if(!state->pause)
            {
            event.user.code = EC_IMG_DELAY;

            if(SDL_PushEvent(&event) < 0)
                return -1;
            }

        return 0;
        }
    }

static int do_delay(state_t *state)
    {
    state->timer_img = SDL_AddTimer(state->ms_img, delay_image, state);

    return 0;
    }

static int do_show_frame(state_t *state)
    {
    SDL_Event event;

    if(state->timer_frame != NULL)
        {
        SDL_RemoveTimer(state->timer_frame);
        state->timer_frame = NULL;
        }

    if(state->next)
        state->framei = state->framec;

    if(state->pause)
        {
        update_image(state);
        return 0;
        }

    if(state->framei < state->framec)
        {
        Uint32 t = SDL_GetTicks();

        update_image(state);

        if(0 == state->framei)
            {
            state->ticks = SDL_GetTicks();
            state->show_ticks = state->ticks - t;
            }
        else
            state->show_ticks += SDL_GetTicks() - t;

        event.user.code = EC_NEXT_FRAME;
        }
    else
        event.user.code = EC_DISPLAY;

    event.user.type = SDL_USEREVENT;
    event.user.data1 = state;
    event.user.data2 = NULL;

    return SDL_PushEvent(&event);
    }

static int do_next_frame(state_t *state)
    {
    SDL_Event event;
    Uint32 next, now;

    ++state->framei;

    next = state->ticks + state->duration * state->framei / state->framec - state->show_ticks / state->framei;
    now = SDL_GetTicks();

    if(next > now)
        {
        next -= now;

        if(next >= 150)
            {
            state->ms_frame = next;
            state->timer_frame = SDL_AddTimer(state->ms_frame, delay_frame, state);

            return 0;
            }
        else
            SDL_Delay(next);
        }

    event.user.type = SDL_USEREVENT;
    event.user.code = EC_SHOW_FRAME;
    event.user.data1 = state;
    event.user.data2 = NULL;

    return SDL_PushEvent(&event);
    }

static int do_continue(state_t *state)
    {
    SDL_Event event;

    if(state->timer_img != NULL)
        {
        SDL_RemoveTimer(state->timer_img);
        state->timer_img = NULL;
        }

    if(state->stop)
        {
        event.type = SDL_QUIT;

        return SDL_PushEvent(&event);
        }
    else if(!state->loader && !state->renderer)
        {
        event.user.type = SDL_USEREVENT;
        event.user.data1 = state;
        event.user.data2 = NULL;

        if(state->framei < state->framec || state->pause)
            event.user.code = EC_DISPLAY;
        else
            event.user.code = EC_RENDER;

        return SDL_PushEvent(&event);
        }
    else
        return 0;
    }

static int do_pause(state_t *state)
    {
    SDL_Event event;

    update_image(state);

    if(state->pause)
        {
        state->pause_ticks = SDL_GetTicks();
        stop_timers(state);
        }
    else
        {
        event.user.type = SDL_USEREVENT;
        event.user.code = EC_CONTINUE;
        event.user.data1 = state;
        event.user.data2 = NULL;

        if(state->framei < state->framec)
            state->ticks += SDL_GetTicks() - state->pause_ticks;

        return SDL_PushEvent(&event);
        }

    return 0;
    }

static int do_next_or_back(state_t *state)
    {
    SDL_Event event;

    if(NULL == state->timer_img  && NULL == state->timer_frame && !state->pause)
        return 0;

    stop_timers(state);

    event.user.type = SDL_USEREVENT;
    event.user.code = EC_CONTINUE;
    event.user.data1 = state;
    event.user.data2 = NULL;

    return SDL_PushEvent(&event);
    }

static int load_image(void *data)
    {
    state_t *state = data;
    SDL_Event event;

    if(load_img2(state) < 0)
        {
        fprintf(stderr, "Bad image: '%s': %s\n", state->imgv[state->imgi], IMG_GetError());

        event.user.type = SDL_USEREVENT;
        event.user.code = EC_IMG_FAILED;
        event.user.data1 = state;
        event.user.data2 = NULL;

        return SDL_PushEvent(&event);
        }

    event.user.type = SDL_USEREVENT;
    event.user.code = EC_IMG_READY;
    event.user.data1 = state;
    event.user.data2 = NULL;

    return SDL_PushEvent(&event);
    }

static int load_img2(state_t *state)
    {
    SDL_Surface *tmp1 = NULL, *tmp2 = NULL;
    SDL_Rect r;
    int w, h;

    w = (state->load_w + 7) & ~7;
    h = (state->load_h + 7) & ~7;

    state->img2 = image_realloc(state->img2, w, h);

    r.x = 0;
    r.y = 0;
    r.w = state->load_w;
    r.h = state->load_h;
    SDL_SetClipRect(state->img2, &r);

    tmp1 = IMG_Load(state->imgv[state->imgi]);
    if(NULL == tmp1)
        {
        fprintf(stderr, "Bad image: '%s': %s\n", state->imgv[state->imgi], IMG_GetError());
        return -1;
        }

    tmp2 = image_convert(tmp1, state->img2->format);

    if(NULL == state->blur || tmp1->w * state->load_h == state->load_w * tmp1->h)
        image_fit(state->img2, tmp2, 0, 0, 0);
    else
        image_fill(state->img2, tmp2, state->blur);

    SDL_FreeSurface(tmp2);
    SDL_FreeSurface(tmp1);

    return 0;
    }

static int render_frames(void *data)
    {
    state_t *state = data;
    SDL_Event event;

    make_frames(state->framev, state->framec, state->img1, state->img2,
        state->mixv[state->mixi], state->blend, &state->stop);

    event.user.type = SDL_USEREVENT;
    event.user.code = EC_FRAMES_READY;
    event.user.data1 = state;
    event.user.data2 = NULL;

    return SDL_PushEvent(&event);
    }

static Uint32 delay_image(Uint32 interval, void *data)
    {
    state_t *state = data;
    SDL_Event event;

    (void)interval; /* Unused */

    event.user.type = SDL_USEREVENT;
    event.user.code = EC_CONTINUE;
    event.user.data1 = state;
    event.user.data2 = NULL;

    SDL_PushEvent(&event);

    return 0;
    }

static Uint32 delay_frame(Uint32 interval, void *data)
    {
    state_t *state = data;
    SDL_Event event;

    (void)interval; /* Unused */

    event.user.type = SDL_USEREVENT;
    event.user.code = EC_SHOW_FRAME;
    event.user.data1 = state;
    event.user.data2 = NULL;

    SDL_PushEvent(&event);

    return 0;
    }

static void stop_timers(state_t *state)
    {
    if(state->timer_img != NULL)
        {
        SDL_RemoveTimer(state->timer_img);
        state->timer_img = NULL;
        }

    if(state->timer_frame != NULL)
        {
        SDL_RemoveTimer(state->timer_frame);
        state->timer_frame = NULL;
        }
    }

static frei0r_plugin_t **select_mixers_by_test(frei0r_list_t *fl)
    {
    frei0r_plugin_t **sel;
    int count = 0;
    int i;

    sel = e_malloc((fl->mixer_count + 1) * sizeof(*sel));

    for(i = 0; i < fl->count; ++i)
        {
        if(test_effect(fl->fp[i]))
            {
            sel[count] = fl->fp[i];
            ++count;
            }
        }

    sel[count] = NULL;

    return sel;
    }

static frei0r_plugin_t **select_mixers_by_list(frei0r_list_t *fl, char **list)
    {
    frei0r_plugin_t **sel;
    int count;
    int i;
    int ok;

    for(count = 0; list[count] != NULL; ++count) {}

    sel = e_malloc((count * fl->count + 1) * sizeof(*sel));

    count = 0;
    while(*list != NULL)
        {
        ok = 0;

        i = -1;
        while((i = frei0r_match(fl, *list, i + 1)) < fl->count)
            {
            if(!is_mixer(fl->fp[i]))
                {
                fprintf(stderr, "Plugin skipped (not transition): '%s'\n",
                    fl->fp[i]->info.name);
                continue;
                }

            sel[count] = fl->fp[i];
            ++count;
            ok = 1;
            }

        if(!ok)
            fprintf(stderr, "Missing transition: '%s'\n", *list);

        ++list;
        }

    sel[count] = NULL;

    return sel;
    }

static void shuffle_mixers(frei0r_plugin_t **sel)
    {
    int count = 0;
    int i, r;
    frei0r_plugin_t *tmp;

    while(sel[count] != NULL)
        ++count;

    srand((unsigned)time(NULL));

    for(i = 0; i < count - 1; ++i)
        {
        r = i + rand() % (count - i);
        tmp = sel[i];
        sel[i] = sel[r];
        sel[r] = tmp;
        }
    }

static int update_image(state_t *state)
    {
    SDL_Surface *icon = state->pause ? state->icon_pause : NULL;

    if(state->framei < state->framec)
        return show_image(state->screen, state->framev[state->framei], 1, icon);
    else if(state->img1)
        return show_image(state->screen, state->img1, 0, icon);

    return 0;
    }

static int show_image(SDL_Surface *screen, SDL_Surface *image, int fast, SDL_Surface *icon)
    {
    int dw, dh;

    dw = image->w - screen->w;
    dh = image->h - screen->h;

    if(dw >= 0 && dw < 8 && dh >= 0 && dh < 8)
        {
        if(image_copy(screen, image) < 0)
            return -1;
        }
    else
        {
        SDL_LockSurface(screen);
        image_fit(screen, image, 0, 0, fast);
        SDL_UnlockSurface(screen);
        }

    if(icon)
        SDL_BlitSurface(icon, NULL, screen, NULL);

    if(SDL_Flip(screen) < 0)
        {
        fprintf(stderr, "SDL_Flip failed: %s\n", SDL_GetError());
        return -1;
        }

    return 0;
    }

static void list_mixers(void)
    {
    frei0r_list_t *plugins;
    int i, j;

    plugins = frei0r_get_plugins();

    if(NULL == plugins)
        {
        fprintf(stderr, "No transitions!\n");
        return;
        }

    for(i = 0; i < plugins->count; ++i)
        {
        if(!test_effect(plugins->fp[i]))
            continue;

        fputs(plugins->fp[i]->info.name, stdout);

        switch(plugins->fp[i]->info.color_model)
            {
        case F0R_COLOR_MODEL_BGRA8888:
            fputs("\t[BGRA]", stdout);
            break;

        case F0R_COLOR_MODEL_RGBA8888:
            fputs("\t[RGBA]", stdout);
            break;

        case F0R_COLOR_MODEL_PACKED32:
            fputs("\t[packed]", stdout);
            break;

        default:
            fputs("\t[UNKNOWN]", stdout);
            break;
            }

        if(plugins->fp[i]->info.num_params > 0)
            {
            fputs("\t(", stdout);
            for(j = 0; j < plugins->fp[i]->info.num_params; ++j)
                {
                if(j > 0)
                    putchar(',');
                fputs(plugins->fp[i]->param_info[j].name, stdout);
                }
            putchar(')');
            }

        putchar('\n');
        }

    frei0r_free_plugins(plugins);
    }

static void progress_tty(unsigned i, unsigned n)
    {
    if(isatty(2))
        {
        fprintf(stderr, "%u/%u\r", i, n);
        if(i == n)
            fprintf(stderr, "\n");
        }
    }

