/*
This file is a part of sleid0r -- slideshow viewer/generator.
Copyright (C) 2013-2014 Vadim Druzhin <cdslow@mail.ru>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <glob.h>
#include <SDL.h>
#include <fnmatch.h>
#include "config.h"
#include "err.h"
#include "mod_frei0r.h"

static char const *Frei0rMask[] =
    {
    "~/.frei0r-1/" CONFIG_INSTALL_LIBDIR "/*.so",
    "~/.frei0r-1/" CONFIG_INSTALL_LIBDIR "/*/*.so",
    "/usr/local/" CONFIG_INSTALL_LIBDIR "/frei0r-1/*.so",
    "/usr/local/" CONFIG_INSTALL_LIBDIR "/frei0r-1/*/*.so",
    "/usr/" CONFIG_INSTALL_LIBDIR "/frei0r-1/*.so",
    "/usr/" CONFIG_INSTALL_LIBDIR "/frei0r-1/*/*.so",
    NULL
    };

typedef void void_f_t(void);
static void_f_t *load_function(void *dll, char const *name, char const *path);

frei0r_list_t *frei0r_get_plugins(void)
    {
    frei0r_list_t *fl;
    glob_t g;
    unsigned i;

    frei0r_glob_all(&g);

    fl = e_malloc(sizeof(*fl) + g.gl_pathc * sizeof(*fl->fp));

    fl->count = 0;
    fl->mixer_count = 0;
    fl->filter_count = 0;
    fl->fp = (frei0r_plugin_t **)(fl + 1);

    for(i = 0; i < g.gl_pathc; ++i)
        {
        fl->fp[fl->count] = frei0r_load(g.gl_pathv[i]);
        if(NULL == fl->fp[fl->count])
            continue;

        if(F0R_PLUGIN_TYPE_FILTER == fl->fp[fl->count]->info.plugin_type)
            {
            ++fl->filter_count;
            ++fl->count;
            }
        else if(F0R_PLUGIN_TYPE_MIXER2 == fl->fp[fl->count]->info.plugin_type)
            {
            ++fl->mixer_count;
            ++fl->count;
            }
        else
            frei0r_close(fl->fp[fl->count]);
        }

    globfree(&g);

    return fl;
    }

void frei0r_free_plugins(frei0r_list_t *fl)
    {
    int i;

    for(i = 0; i < fl->count; ++i)
        frei0r_close(fl->fp[i]);

    free(fl);
    }

void frei0r_glob_all(glob_t *g)
    {
    char const **mask;
    int g_append = 0;
    int ret;

    for(mask = Frei0rMask; *mask != NULL; ++mask)
        {
        ret = glob(*mask, g_append | GLOB_TILDE, NULL, g);
        if(ret != 0 && ret != GLOB_NOMATCH)
            {
            perror("frei0r_glob_all");
            exit(EXIT_FAILURE);
            }
        g_append = GLOB_APPEND;
        }
    }

frei0r_plugin_t *frei0r_load(char *path)
    {
    frei0r_plugin_t *fp = NULL;
    int i;

    do
        {
        fp = e_malloc(sizeof(*fp));

        memset(fp, 0, sizeof(*fp));

        fp->dll = SDL_LoadObject(path);
        if(NULL == fp->dll)
            {
            printf("%s: %s\n", path, SDL_GetError());
            break;
            }
    
        fp->f_init = (f0r_init_t *)load_function(fp->dll, "f0r_init", path);
        if(NULL == fp->f_init)
            break;

        fp->f_deinit = load_function(fp->dll, "f0r_deinit", path);
        if(NULL == fp->f_deinit)
            break;

        fp->f_get_plugin_info = (f0r_get_plugin_info_t *)load_function(fp->dll, "f0r_get_plugin_info", path);
        if(NULL == fp->f_get_plugin_info)
            break;

        fp->f_construct = (f0r_construct_t *)load_function(fp->dll, "f0r_construct", path);
        if(NULL == fp->f_construct)
            break;

        fp->f_destruct = (f0r_destruct_t *)load_function(fp->dll, "f0r_destruct", path);
        if(NULL == fp->f_destruct)
            break;

        fp->f_get_param_info = (f0r_get_param_info_t *)load_function(fp->dll, "f0r_get_param_info", NULL);
        fp->f_set_param_value = (f0r_set_param_value_t *)load_function(fp->dll, "f0r_set_param_value", NULL);
        fp->f_get_param_value = (f0r_get_param_value_t *)load_function(fp->dll, "f0r_get_param_value", NULL);
        fp->f_update = (f0r_update_t *)load_function(fp->dll, "f0r_update", NULL);
        fp->f_update2 = (f0r_update2_t *)load_function(fp->dll, "f0r_update2", NULL);

        fp->init = fp->f_init();
        fp->f_get_plugin_info(&fp->info);

        if(fp->info.frei0r_version != FREI0R_MAJOR_VERSION)
            {
            fprintf(stderr, "%s: wrong frei0r version %d\n", path, fp->info.frei0r_version );
            break;
            }

        if(fp->info.num_params > 0)
            {
            if(NULL == fp->f_get_param_info)
                {
                fprintf(stderr, "%s: missing %s()\n", path, "f0r_get_param_info");
                break;
                }
            else
                {
                fp->param_info = e_malloc(fp->info.num_params * sizeof(*fp->param_info));

                for(i = 0; i < fp->info.num_params; ++i)
                    fp->f_get_param_info(fp->param_info + i, i);
                }
            }

        return fp;
        }
    while(0);

    frei0r_close(fp);

    return NULL;
    }

void frei0r_close(frei0r_plugin_t *fp)
    {
    if(fp != NULL)
        {
        if(fp->param_info != NULL)
            free(fp->param_info);

        if(fp->init)
            fp->f_deinit();

        if(fp->dll != NULL)
            SDL_UnloadObject(fp->dll);

        free(fp);
        }
    }

static void_f_t *load_function(void *dll, char const *name, char const *path)
    {
    void (*ret)();

    ret = (void_f_t *)(uintptr_t)SDL_LoadFunction(dll, name);
    if(NULL == ret && NULL != path)
        fprintf(stderr, "%s: missing %s()\n", path, name);

    return ret;
    }

frei0r_plugin_t *frei0r_search(frei0r_list_t const *fl, char const *name)
    {
    int i;

    for(i = 0; i < fl->count; ++i)
        {
        if(strcmp(fl->fp[i]->info.name, name) == 0)
            return fl->fp[i];
        }

    return NULL;
    }

int frei0r_match(frei0r_list_t const *fl, char const *mask, int i)
    {
    for(; i < fl->count; ++i)
        {
        if(fnmatch(mask, fl->fp[i]->info.name, 0) == 0)
            break;
        }

    return i;
    }
