/*
This file is a part of sleid0r -- slideshow viewer/generator.
Copyright (C) 2013-2015 Vadim Druzhin <cdslow@mail.ru>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include "err.h"
#include "options.h"

typedef struct opt_map_s
    {
    char *opt;
    int id;
    } opt_map_t;

enum OPT_ID
    {
    OPT_SORT = 0,
    OPT_BLACK,
    OPT_COLORED,
    OPT_DELAY,
    OPT_EFFECTS,
    OPT_FULLSCREEN,
    OPT_ALLSCREEN,
    OPT_GEOM,
    OPT_HELP,
    OPT_PLAYLIST,
    OPT_LIST,
    OPT_NO_REPEAT,
    OPT_OUT,
    OPT_REPEAT,
    OPT_FPS,
    OPT_SHUFFLE,
    OPT_E_SHUFFLE,
    OPT_DURATION,
    OPT_UNSORTED,
    OPT_E_UNSORTED,
    OPT_D_SHUFFLE,
    OPT_SPACE_NEXT,
    OPT_SPACE_PAUSE,
    OPT_COUNT
    };

static opt_map_t const Options[]=
    {
    {"-a", OPT_SORT},
    {"--sort", OPT_SORT},
    {"-b", OPT_BLACK},
    {"--black", OPT_BLACK},
    {"-c", OPT_COLORED},
    {"--colored", OPT_COLORED},
    {"-d", OPT_DELAY},
    {"--delay", OPT_DELAY},
    {"-e", OPT_EFFECTS},
    {"--effects", OPT_EFFECTS},
    {"-f", OPT_FULLSCREEN},
    {"--full-screen", OPT_FULLSCREEN},
    {"-F", OPT_ALLSCREEN},
    {"--all-screen", OPT_ALLSCREEN},
    {"-g", OPT_GEOM},
    {"--geometry", OPT_GEOM},
    {"-h", OPT_HELP},
    {"--help", OPT_HELP},
    {"-i", OPT_PLAYLIST},
    {"--playlist", OPT_PLAYLIST},
    {"-l", OPT_LIST},
    {"--list", OPT_LIST},
    {"-n", OPT_NO_REPEAT},
    {"--no-repeat", OPT_NO_REPEAT},
    {"-o", OPT_OUT},
    {"--output", OPT_OUT},
    {"-p", OPT_REPEAT},
    {"--repeat", OPT_REPEAT},
    {"-r", OPT_FPS},
    {"--fps", OPT_FPS},
    {"-s", OPT_SHUFFLE},
    {"--shuffle", OPT_SHUFFLE},
    {"-S", OPT_E_SHUFFLE},
    {"--shuffle-effects", OPT_E_SHUFFLE},
    {"-t", OPT_DURATION},
    {"--duration", OPT_DURATION},
    {"-u", OPT_UNSORTED},
    {"--unsorted", OPT_UNSORTED},
    {"-U", OPT_E_UNSORTED},
    {"--unsorted-effects", OPT_E_UNSORTED},
    {"-y", OPT_D_SHUFFLE},
    {"--shuffle-dirs", OPT_D_SHUFFLE},
    {"--space-is-next", OPT_SPACE_NEXT},
    {"--space-is-pause", OPT_SPACE_PAUSE},
    {NULL, -1}
    };

static char const *UsageText[]=
    {
    "sleid0r version 0.4 GPL 3+\n",
    "Copyright (C) 2013-2015 Vadim Druzhin <cdslow@mail.ru>\n"
    "\n"
    "For licensing details please visit <http://www.gnu.org/licenses/>.\n"
    "\n"
    "Display (or generate) a slideshow\n"
    "using the frei0r mixer effects as the transitions.\n"
    "\n"
    "Call: sleid0r [options] <images>\n"
    "\n",
    NULL
    };

static char const *ConfigText[]=
    {
    "Configuration files:\n"
    "\n"
    "$XDG_CONFIG_HOME/sleid0r/sleid0r.cfg\n"
    "$HOME/.config/sleid0r/sleid0r.cfg\n"
    "/etc/sleid0r.cfg\n"
    "\n"
    "First available file will be used.\n"
    "Options must be stored exactly as presented on command line,\n"
    "but may be splitted across several lines.\n"
    "\n",
    "Control keys:\n"
    "\n"
    "ESC,q - quit;\n"
    "f - toggle full screen;\n"
    "F - toggle all screen;\n"
    "p - pause slideshow;\n"
    "c - continue slideshow;\n"
    "PAUSE - toggle pause;\n"
    "PAGEDOWN, n - show next image;\n"
    "PAGEUP, BACKSPACE, SHIFT-SPACE, b - show previous image;\n"
    "SPACE - show next image or toggle pause.\n"
    "\n",
    "Examples:\n"
    "\n"
    "sleid0r .\n"
    "  - permanently display a slideshow using all images\n"
    "    from the current directory (including subdirectories).\n"
    "\n"
    "sleid0r -e 'wipe-*' -o ~/frames/ *.jpg\n"
    "  - generate a set of BMP files from JPEG images\n"
    "    using all wipe transitions.\n"
    "\n"
    "sleid0r -i images.lst -g 768x576 -r 25 -o - | \\\n"
    "ffmpeg -f rawvideo -video_size 768x576 -r 25 -pix_fmt rgba -i pipe: -target dvd -aspect 4:3 show.mpg\n"
    "  - generate the MPEG2 slideshow 'show.mpg'\n"
    "    from the images listed in 'images.lst'.\n"
    "\n",
    NULL
    };

static opt_map_t const OptsDesc[]=
    {
    {" - display this help;\n", OPT_HELP},
    {" - list available transition effects;\n", OPT_LIST},
    {" - alphabeticaly sort the images (default);\n", OPT_SORT},
    {" - black background;\n", OPT_BLACK},
    {" - colored background (default);\n", OPT_COLORED},
    {" - shuffle images;\n", OPT_SHUFFLE},
    {" - shuffle directories;\n", OPT_D_SHUFFLE},
    {" - shuffle transition effects;\n", OPT_E_SHUFFLE},
    {" - use the list of images as is;\n", OPT_UNSORTED},
    {" - use the list of transitions as is (default);\n", OPT_E_UNSORTED},
    {" - permanently repeat all images (default);\n", OPT_REPEAT},
    {" - display images only once and exit;\n", OPT_NO_REPEAT},
    {" <seconds> - image change delay (3 seconds by default);\n", OPT_DELAY},
    {" <seconds> - transition duration (1 second by default);\n", OPT_DURATION},
    {" <number>[/<number>] - transition frame rate (30 fps by default);\n", OPT_FPS},
    {" <list,...> - comma delimited list of transition effects;\n", OPT_EFFECTS},
    {" [monitor] - run in full screen (on given monitor, starting from 0);\n", OPT_FULLSCREEN},
    {" - run in full screen (over all monitors);\n", OPT_ALLSCREEN},
    {" <width>x<height> - window geometry in pixels;\n", OPT_GEOM},
    {" <filename> - read the list of images from a file\n"
     "    (use '-' for standard input);\n", OPT_PLAYLIST},
    {" <path> - output the slideshow as sequence of images at given path\n"
     "    (use '-' for standard output);\n", OPT_OUT},
    {" - use SPACE key for display next image (default);\n", OPT_SPACE_NEXT},
    {" - use SPACE key for toggle pause.\n", OPT_SPACE_PAUSE},
    {NULL, -1}
    };

static int opt_match(opt_map_t const *map, char const *opt);
static int parse_seconds(double *sec, char **argv);
static int parse_fps(double *fps, char **argv);
static int parse_geometry(int *w, int *h, char **argv);
static char *parse_display(char **argv);
static char **parse_list(char **argv);
static char **parse_file(char **argv);
static void shift_argv(char **argv);
static char *check_path(char const *a, char const *b);
static char *join_path(char const *a, char const *b);
static char *read_file(FILE *in);
static char *opt_cfg_path(char const *cfg_name);
static char **read_words(FILE *in);
static char **read_lines(FILE *in);
static void free_lines(char **lines);
static void opt_help(opt_map_t const *desc, opt_map_t const *map);

void init_params(param_t *params)
    {
    params->help = 0;
    params->list = 0;
    params->sort = 1;
    params->shuffle = 0;
    params->d_shuffle = 0;
    params->e_shuffle = 0;
    params->black = 0;
    params->repeat = 1;
    params->delay = 3.0;
    params->duration = 1.0;
    params->fps = 30.0;
    params->selection = NULL;
    params->fs = 0;
    params->fs_id = "0";
    params->w = 640;
    params->h = 480;
    params->playlist = NULL;
    params->prefix = NULL;
    params->space_is_pause = 0;
    }

void free_params(param_t *params)
    {
    if(params->selection != NULL)
        free(params->selection);

    if(params->playlist != NULL)
        free_lines(params->playlist);
    }

int read_config(param_t *params, char const *name)
    {
    char *cfg_name;
    FILE *cfg_file;
    char **cfgv;
    int ret = 1;

    cfg_name = opt_cfg_path(name);

    if(cfg_name != NULL)
        {
        cfg_file = fopen(cfg_name, "r");

        if(NULL == cfg_file)
            perror(cfg_name);
        else
            {
            cfgv = read_words(cfg_file);

            fclose(cfg_file);

            ret = parse_options(params, cfgv);

            if(cfgv[0] != NULL)
                {
                fprintf(stderr, "Bad option in '%s': '%s'\n", cfg_name, cfgv[0]);
                ret = 0;
                }

            free(cfgv);
            }
        free(cfg_name);
        }

    return ret;
    }

int parse_options(param_t *params, char **argv)
    {
    while(*argv != NULL)
        {
        switch(opt_match(Options, *argv))
            {
        case OPT_LIST:
            params->list = 1;
            break;

        case OPT_HELP:
            params->help = 1;
            break;

        case OPT_SORT:
            params->sort = 1;
            params->shuffle = 0;
            params->d_shuffle = 0;
            break;

        case OPT_SHUFFLE:
            params->sort = 0;
            params->shuffle = 1;
            params->d_shuffle = 0;
            break;

        case OPT_UNSORTED:
            params->sort = 0;
            params->shuffle = 0;
            params->d_shuffle = 0;
            break;

        case OPT_D_SHUFFLE:
            params->d_shuffle = 1;
            break;

        case OPT_E_SHUFFLE:
            params->e_shuffle = 1;
            break;

        case OPT_E_UNSORTED:
            params->shuffle = 0;
            break;

        case OPT_BLACK:
            params->black = 1;
            break;

        case OPT_COLORED:
            params->black = 0;
            break;

        case OPT_REPEAT:
            params->repeat = 1;
            break;

        case OPT_NO_REPEAT:
            params->repeat = 0;
            break;

        case OPT_DELAY:
            if(!parse_seconds(&params->delay, argv))
                return 0;
            break;

        case OPT_DURATION:
            if(!parse_seconds(&params->duration, argv))
                return 0;
            break;

        case OPT_FPS:
            if(!parse_fps(&params->fps, argv))
                return 0;
            break;

        case OPT_EFFECTS:
            params->selection = parse_list(argv);
            if(NULL == params->selection)
                return 0;
            break;

        case OPT_FULLSCREEN:
            params->fs = 1;
            params->fs_id = parse_display(argv);
            break;

        case OPT_ALLSCREEN:
            params->fs = 2;
            break;

        case OPT_GEOM:
            if(!parse_geometry(&params->w, &params->h, argv))
                return 0;
            params->fs = 0;
            break;

        case OPT_PLAYLIST:
            params->playlist = parse_file(argv);
            if(NULL == params->playlist)
                return 0;
            break;

        case OPT_OUT:
            if(NULL == argv[1])
                {
                fprintf(stderr, "Missing path after '%s'\n", argv[0]);
                return 0;
                }
            params->prefix = argv[1];
            shift_argv(argv);
            break;

        case OPT_SPACE_NEXT:
            params->space_is_pause = 0;
            break;

        case OPT_SPACE_PAUSE:
            params->space_is_pause = 1;
            break;

        default:
            return 1;
            }
        shift_argv(argv);
        }

    return 1;
    }

static void shift_argv(char **argv)
    {
    while(*argv != NULL)
        {
        argv[0] = argv[1];
        ++argv;
        }
    }

static int parse_seconds(double *sec, char **argv)
    {
    char *tmp;

    if(NULL == argv[1])
        {
        fprintf(stderr, "Missing number of seconds after '%s'\n", argv[0]);
        return 0;
        }

    *sec = strtod(argv[1], &tmp);
    if(tmp == argv[1] || *tmp != 0 || *sec < 0.0)
        {
        fprintf(stderr, "Bad number of seconds: '%s'\n", argv[1]);
        return 0;
        }

    shift_argv(argv);

    return 1;
    }

static int parse_fps(double *fps, char **argv)
    {
    char *tmp;
    double div;

    if(NULL == argv[1])
        {
        fprintf(stderr, "Missing frame rate after '%s'\n", argv[0]);
        return 0;
        }

    *fps = strtod(argv[1], &tmp);
    if(tmp == argv[1] || (*tmp != 0 && *tmp != '/') || *fps < 0.0)
        {
        fprintf(stderr, "Bad frame rate: '%s'\n", argv[1]);
        return 0;
        }

    if('/' == *tmp)
        {
        div = strtod(tmp + 1, &tmp);
        if(*tmp != 0 || div <= 0.0)
            {
            fprintf(stderr, "Bad frame rate divisor: '%s'\n", argv[1]);
            return 0;
            }
        *fps = *fps / div;
        }

    shift_argv(argv);

    return 1;
    }

static int parse_geometry(int *w, int *h, char **argv)
    {
    char *tmp;

    if(NULL == argv[1])
        {
        fprintf(stderr, "Missing geometry after '%s'\n", argv[0]);
        return 0;
        }

    *w = (int)strtol(argv[1], &tmp, 10);
    if(tmp == argv[1] || (*tmp != 0 && *tmp != 'x') || *w <= 0)
        {
        fprintf(stderr, "Bad geometry width: '%s'\n", argv[1]);
        return 0;
        }

    if('x' == *tmp)
        {
        *h = (int)strtol(tmp + 1, &tmp, 10);
        if(*tmp != 0 || *h <= 0)
            {
            fprintf(stderr, "Bad geometry height: '%s'\n", argv[1]);
            return 0;
            }
        }

    shift_argv(argv);

    return 1;
    }

static char *parse_display(char **argv)
    {
    char *ret = "0";
    int i;

    if(NULL == argv[1])
        return ret;

    for(i = 0; argv[1][i] != 0; ++i)
        {
        if(!isdigit(argv[1][i]))
            return ret;
        }

    ret = argv[1];

    shift_argv(argv);

    return ret;
    }

static char **parse_list(char **argv)
    {
    char **list = NULL;
    int count;
    char *s;


    if(NULL == argv[1] || 0 == argv[1][0])
        {
        fprintf(stderr, "Missing list after '%s'\n", argv[0]);
        return NULL;
        }

    count = 1;
    s = argv[1] - 1;
    while((s = strchr(s + 1, ',')) != NULL)
        ++count;

    list = e_malloc((count + 1) * sizeof(*list) + strlen(argv[1]) + 1);

    list[0] = (char *)(list + count + 1);

    strcpy(list[0], argv[1]);

    count = 0;
    while((s = strchr(list[count], ',')) != NULL)
        {
        *s = 0;
        ++count;
        list[count] = s + 1;
        }

    list[count + 1] = NULL;

    shift_argv(argv);

    return list;
    }

static char **parse_file(char **argv)
    {
    char **list;
    FILE *in;

    if(NULL == argv[1] || 0 == argv[1][0])
        {
        fprintf(stderr, "Missing file name after '%s'\n", argv[0]);
        return NULL;
        }

    if(strcmp(argv[1], "-") == 0)
        in = stdin;
    else
        in = fopen(argv[1], "r");

    if(NULL == in)
        {
        fprintf(stderr, "Can't open file '%s'\n", argv[1]);
        perror(NULL);
        return NULL;
        }

    list = read_lines(in);

    if(in != stdin)
        fclose(in);

    shift_argv(argv);

    return list;
    }

static int opt_match(opt_map_t const *map, char const *opt)
    {
    while(NULL != map->opt)
        {
        if(0 == strcmp(map->opt, opt))
            return map->id;
        ++map;
        }

    return -1;
    }

static char **read_lines(FILE *in)
    {
    char **lines = NULL;
    int count = 0;
    char buf[1024];
    int lines_nl = 1;
    int pos;
    int buf_nl;

    lines = e_malloc(sizeof(*lines));

    lines[count] = NULL;

    while(fgets(buf, sizeof(buf), in) != NULL)
        {
        pos = strlen(buf) - 1;

        if(buf[pos] == '\n')
            {
            buf_nl = 1;
            while(pos >= 0 && (buf[pos] == '\n' || buf[pos] == '\r'))
                {
                buf[pos] = 0;
                --pos;
                }
            }
        else
            buf_nl = 0;

        if(pos < 0)
            continue;

        if(lines_nl)
            {
            lines = e_realloc(lines, (count + 2) * sizeof(*lines));

            lines[count] = e_strdup(buf);
            }
        else
            {
            lines[count] = e_realloc(lines[count], strlen(lines[count]) + strlen(buf) + 1);

            strcat(lines[count], buf);
            }

        lines_nl = buf_nl;

        if(lines_nl)
            {
            ++count;
            lines[count] = NULL;
            }
        }

    if(ferror(in))
        perror("read_lines");

    if(!lines_nl)
        {
        ++count;
        lines[count] = NULL;
        }

    return lines;
    }

static void free_lines(char **lines)
    {
    int i;

    if(lines != NULL)
        {
        for(i = 0; lines[i] != NULL; ++i)
            free(lines[i]);

        free(lines);
        }
    }

static char *opt_cfg_path(char const *cfg_name)
    {
    char *dir;
    char *path;

    dir = getenv("XDG_CONFIG_HOME");
    if(dir != NULL)
        {
        dir = join_path(dir, "sleid0r");

        path = check_path(dir, cfg_name);

        free(dir);

        if(path !=  NULL)
            return path;
        }

    dir = getenv("HOME");
    if(dir != NULL)
        {
        dir = join_path(dir, ".config/sleid0r");

        path = check_path(dir, cfg_name);

        free(dir);

        if(path !=  NULL)
            return path;
        }

    path = check_path("/etc", cfg_name);

    return path;
    }

static char *check_path(char const *a, char const *b)
    {
    char *path = join_path(a, b);

    if(access(path, F_OK) != 0)
        {
        free(path);
        path = NULL;
        }

    return path;
    }

static char *join_path(char const *a, char const *b)
    {
    char *path = e_malloc(strlen(a) + 1 + strlen(b) + 1);

    strcat(strcat(strcpy(path, a), "/"), b);

    return path;
    }

static char **read_words(FILE *in)
    {
    char **words = NULL;
    char *buf = NULL;
    int len;
    int wc;
    int in_space, in_quotes;
    char *s;
    char *dst;

    buf = read_file(in);
    if(NULL == buf)
        {
        perror("read_words");
        return NULL;
        }

    wc = 0;
    in_space = 1;
    in_quotes = 0;

    for(s = buf; *s != 0; ++s)
        {
        if(in_quotes)
            {
            if(in_quotes == *s)
                in_quotes = 0;
            continue;
            }

        if(isspace(*s))
            in_space = 1;
        else
            {
            if(in_space)
                ++wc;

            in_space = 0;

            if('"' == *s || '\'' == *s)
                in_quotes = *s;
            }
        }

    len = s - buf;

    words = e_malloc((wc + 1) * sizeof(*words) + len + 1);
    dst = (char *)(words + wc + 1);

    wc = 0;
    in_space = 1;
    in_quotes = 0;

    for(s = buf; *s != 0; ++s)
        {
        if(in_quotes)
            {
            if(in_quotes == *s)
                in_quotes = 0;
            else
                *dst++ = *s;

            continue;
            }

        if(isspace(*s))
            {
            if(!in_space)
                *dst++ = 0;

            in_space = 1;
            }
        else
            {
            if(in_space)
                words[wc++] = dst;

            in_space = 0;

            if('"' == *s || '\'' == *s)
                in_quotes = *s;
            else
                *dst++ = *s;
            }
        }

    *dst = 0;

    words[wc] = NULL;

    free(buf);

    return words;
    }

static char *read_file(FILE *in)
    {
    char *buf = NULL;
    long flen;

    if(fseek(in, 0l, SEEK_END) < 0)
        return NULL;

    flen = ftell(in);
    if(-1l == flen)
        return NULL;

    rewind(in);

    buf = e_malloc(flen + 1);

    if((long)fread(buf, 1, flen, in) != flen)
        {
        free(buf);
        return NULL;
        }

    buf[flen] = 0;

    return buf;
    }

void usage(void)
    {
    char const **txt;

    for(txt = UsageText; *txt != NULL; ++txt)
        fputs(*txt, stdout);

    puts("Options:\n");
    opt_help(OptsDesc, Options);
    putchar('\n');

    for(txt = ConfigText; *txt != NULL; ++txt)
        fputs(*txt, stdout);
    }

static void opt_help(opt_map_t const *desc, opt_map_t const *map)
    {
    int count[OPT_COUNT];
    opt_map_t const *m;
    int i;
    
    memset(count, 0, sizeof(count));
    for(m = map; m->id >= 0 && m->id < OPT_COUNT; ++m)
        ++count[m->id];

    for(i = 0; i < OPT_COUNT; ++i)
        {
        if(0 == count[i])
            continue;
        for(m = map; NULL != m->opt; ++m)
            {
            if(m->id == i)
                {
                fputs(m->opt, stdout);
                --count[i];
                if(count[i] > 0)
                    putchar('|');
                }
            }
        for(m = desc; NULL != m->opt; ++m)
            {
            if(m->id == i)
                fputs(m->opt, stdout);
            }
        }
    }
