/*
This file is a part of sleid0r -- slideshow viewer/generator.
Copyright (C) 2013-2015 Vadim Druzhin <cdslow@mail.ru>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <SDL.h>
#include "mod_frei0r.h"
#include "images.h"
#include "err.h"
#include "slide.h"

static void make_frames_smooth(
    SDL_Surface **framev,
    int count,
    SDL_Surface *img1,
    SDL_Surface *img2,
    frei0r_plugin_t *effect,
    int volatile *stop
    );
static void make_frames_blend(
    SDL_Surface **framev,
    int count,
    SDL_Surface *img1,
    SDL_Surface *img2,
    frei0r_plugin_t *effect,
    frei0r_plugin_t *blend,
    int volatile *stop
    );
static void make_frames_static(
    SDL_Surface **framev,
    int count,
    SDL_Surface *img1,
    SDL_Surface *img2,
    frei0r_plugin_t *effect,
    int volatile *stop
    );
static SDL_Surface *make_middle(SDL_Surface *img1, SDL_Surface *img2, frei0r_plugin_t *effect);
static void prepare_effect(
    SDL_Surface **out,
    SDL_Surface **tmp1,
    SDL_Surface **tmp2,
    SDL_Surface *img1,
    SDL_Surface *img2,
    frei0r_plugin_t *effect
    );
static void format_idx(char *num, int num_w, unsigned idx);

void make_frames(
    SDL_Surface **framev,
    int count,
    SDL_Surface *img1,
    SDL_Surface *img2,
    frei0r_plugin_t *effect,
    frei0r_plugin_t *blend,
    int volatile *stop
    )
    {
    SDL_Surface *tmp1 = NULL;
    int i;

    if(img1->w != img2->w || img1->h != img2->h)
        {
        tmp1 = image_realloc(NULL, img2->w, img2->h);
        SDL_SetClipRect(tmp1, &img2->clip_rect);
        image_fit(tmp1, img1, 0, 0, 0);
        img1 = tmp1;
        }

    for(i = 0; i < count && !*stop; ++i)
        framev[i] = image_realloc(framev[i], img2->w, img2->h);

    if(*stop)
        {}
    else if(effect->info.num_params > 0 && F0R_PARAM_DOUBLE == effect->param_info[0].type)
        make_frames_smooth(framev, count, img1, img2, effect, stop);
    else if(blend != NULL)
        make_frames_blend(framev, count, img1, img2, effect, blend, stop);
    else
        make_frames_static(framev, count, img1, img2, effect, stop);

    if(tmp1 != NULL)
        SDL_FreeSurface(tmp1);
    }

static void make_frames_smooth(
    SDL_Surface **framev,
    int count,
    SDL_Surface *img1,
    SDL_Surface *img2,
    frei0r_plugin_t *effect,
    int volatile *stop
    )
    {
    SDL_Surface *out;
    SDL_Surface *tmp1 = NULL, *tmp2 = NULL;
    f0r_instance_t inst = NULL;
    double amount;
    int i;

    prepare_effect(&out, &tmp1, &tmp2, img1, img2, effect);

    if(tmp1 != NULL)
        img1 = tmp1;

    if(tmp2 != NULL)
        img2 = tmp2;

    inst = effect->f_construct(out->w, out->h);

    for(i = 0; i < count && !*stop; ++i)
        {
        amount = (double)(i + 1) / (double)(count + 1);
        effect->f_set_param_value(inst, &amount, 0);
        effect->f_update2(inst, amount,
            img1->pixels, img2->pixels, NULL, out->pixels);
        image_copy(framev[i], out);
        }

    effect->f_destruct(inst);

    if(tmp2 != NULL)
        SDL_FreeSurface(tmp2);

    if(tmp1 != NULL)
        SDL_FreeSurface(tmp1);

    SDL_FreeSurface(out);
    }

static void make_frames_blend(
    SDL_Surface **framev,
    int count,
    SDL_Surface *img1,
    SDL_Surface *img2,
    frei0r_plugin_t *effect,
    frei0r_plugin_t *blend,
    int volatile *stop
    )
    {
    SDL_Surface *middle;
    SDL_Surface *out;
    SDL_Surface *tmp1 = NULL, *tmp2 = NULL, *tmp3 = NULL;
    f0r_instance_t inst = NULL;
    double amount;
    int i;

    prepare_effect(&out, &tmp1, &tmp2, img1, img2, blend);

    if(tmp1 != NULL)
        img1 = tmp1;

    if(tmp2 != NULL)
        img2 = tmp2;

    middle = make_middle(img1, img2, effect);

    if(middle->format->Rmask != out->format->Rmask)
        {
        tmp3 = image_convert(middle, out->format);
        SDL_FreeSurface(middle);
        middle = tmp3;
        }

    inst = blend->f_construct(out->w, out->h);

    for(i = 0; i < count && !*stop; ++i)
        {
        amount = (double)(i + 1) * 2 / (double)(count + 1);
        if(amount <= 1.0)
            {
            blend->f_set_param_value(inst, &amount, 0);
            blend->f_update2(inst, amount,
                img1->pixels, middle->pixels, NULL, out->pixels);
            }
        else
            {
            amount -= 1.0;
            blend->f_set_param_value(inst, &amount, 0);
            blend->f_update2(inst, amount,
                middle->pixels, img2->pixels, NULL, out->pixels);
            }
        image_copy(framev[i], out);
        }

    blend->f_destruct(inst);

    SDL_FreeSurface(middle);

    if(tmp2 != NULL)
        SDL_FreeSurface(tmp2);

    if(tmp1 != NULL)
        SDL_FreeSurface(tmp1);

    SDL_FreeSurface(out);
    }

static void make_frames_static(
    SDL_Surface **framev,
    int count,
    SDL_Surface *img1,
    SDL_Surface *img2,
    frei0r_plugin_t *effect,
    int volatile *stop
    )
    {
    SDL_Surface *middle;
    int i;

    middle = make_middle(img1, img2, effect);

    for(i = 0; i < count && !*stop; ++i)
        image_copy(framev[i], middle);

    SDL_FreeSurface(middle);
    }

static SDL_Surface *make_middle(SDL_Surface *img1, SDL_Surface *img2, frei0r_plugin_t *effect)
    {
    SDL_Surface *middle;
    SDL_Surface *tmp1 = NULL, *tmp2 = NULL;
    f0r_instance_t inst = NULL;

    prepare_effect(&middle, &tmp1, &tmp2, img1, img2, effect);

    if(tmp1 != NULL)
        img1 = tmp1;

    if(tmp2 != NULL)
        img2 = tmp2;

    inst = effect->f_construct(middle->w, middle->h);

    effect->f_update2(inst, 0.5,
        img1->pixels, img2->pixels, NULL, middle->pixels);

    effect->f_destruct(inst);

    if(tmp2 != NULL)
        SDL_FreeSurface(tmp2);

    if(tmp1 != NULL)
        SDL_FreeSurface(tmp1);

    return middle;
    }

static void prepare_effect(
    SDL_Surface **out,
    SDL_Surface **tmp1,
    SDL_Surface **tmp2,
    SDL_Surface *img1,
    SDL_Surface *img2,
    frei0r_plugin_t *effect
    )
    {
    *tmp1 = NULL;
    *tmp2 = NULL;

    if(F0R_COLOR_MODEL_BGRA8888 == effect->info.color_model)
        *out = image_alloc_bgra(img2->w, img2->h);
    else if(F0R_COLOR_MODEL_RGBA8888 == effect->info.color_model)
        *out = image_alloc_rgba(img2->w, img2->h);
    else
        *out = image_realloc(NULL, img2->w, img2->h);

    if(img1->format->Rmask != (*out)->format->Rmask)
        *tmp1 = image_convert(img1, (*out)->format);

    if(img2->format->Rmask != (*out)->format->Rmask)
        *tmp2 = image_convert(img2, (*out)->format);
    }

int is_mixer(frei0r_plugin_t *effect)
    {
    return effect->info.plugin_type == F0R_PLUGIN_TYPE_MIXER2;
    }

int test_effect(frei0r_plugin_t *effect)
    {
    SDL_Surface *test1, *test2, *middle;
    f0r_instance_t inst = NULL;
    double amount = 0.5;
    int ret = 1;

    if(!is_mixer(effect))
        return 0;

    test1 = image_alloc_rgba(16, 16);
    test2 = image_alloc_rgba(16, 16);
    middle = image_alloc_rgba(16, 16);

    image_random(test1);
    image_random(test2);
    image_black(middle);

    inst = effect->f_construct(middle->w, middle->h);

    if(effect->info.num_params > 0 && F0R_PARAM_DOUBLE == effect->param_info[0].type)
        effect->f_set_param_value(inst, &amount, 0);

    effect->f_update2(inst, amount,
        test1->pixels, test2->pixels, NULL, middle->pixels);

    effect->f_destruct(inst);

    if(image_cmp(test1, middle) == 0)
        ret = 0;
    else if(image_cmp(test2, middle) == 0)
        ret = 0;

    SDL_FreeSurface(middle);
    SDL_FreeSurface(test2);
    SDL_FreeSurface(test1);

    return ret;
    }

unsigned write_frames(char *path, char *num, int num_w, unsigned idx, SDL_Surface **framev, int framec)
    {
    int i;

    for(i = 0; i < framec; ++i)
        {
        if(path != NULL)
            {
            format_idx(num, num_w, idx);
            if(SDL_SaveBMP(framev[i], path) < 0)
                {
                fprintf(stderr, "write_frames: %s\n", SDL_GetError());
                return 0;
                }
            }
        else
            {
            if(fwrite(framev[i]->pixels, framev[i]->format->BytesPerPixel,
                    framev[i]->w * framev[i]->h, stdout) != (size_t)framev[i]->w * framev[i]->h)
                {
                perror("write_frames");
                return 0;
                }
            }
        ++idx;
        }

    return idx;
    }

unsigned write_images(char *path, char *num, int num_w, unsigned idx, SDL_Surface *image, int count)
    {
    int i;

    for(i = 0; i < count; ++i)
        {
        if(path != NULL)
            {
            format_idx(num, num_w, idx);
            if(SDL_SaveBMP(image, path) < 0)
                {
                fprintf(stderr, "write_images: %s\n", SDL_GetError());
                return 0;
                }
            }
        else
            {
            if(fwrite(image->pixels, image->format->BytesPerPixel,
                    image->w * image->h, stdout) != (size_t)image->w * image->h)
                {
                perror("write_images");
                return 0;
                }
            }
        ++idx;
        }

    return idx;
    }

static void format_idx(char *num, int num_w, unsigned idx)
    {
    sprintf(num, "%0*u.bmp", num_w, idx);
    }

char *prepare_path(char *prefix, int num_w)
    {
    int len;
    char *path;

    len = strlen(prefix);
    path = e_malloc(len + num_w + 5);
    strcpy(path, prefix);

    return path;
    }

